import copy
import logging
import os
from typing import Optional, Tuple
from dcicutils import ff_utils
from dcicutils.env_utils import foursight_env_name
from dcicutils.function_cache_decorator import function_cache
from dcicutils.misc_utils import find_association
from ...app import app
from .gac import Gac

logging.basicConfig()
logger = logging.getLogger(__name__)


# TODO
# Rationalize this with dcicutil.utils env_utils functions for name versions, normalizations, comparisons.
# Did at least some of this (below, e.g. find_known_env, is_same_env); but see if we can perhaps get away
# from even having to keep these (5) name versions around (i.e. name, short_name, full_name, public_name,
# foursight_name); the UI currently relies on these.
class Envs:

    _DEFAULT_ENV_PLACHOLDER = 'no-default-env'

    def __init__(self, known_envs: list):
        # This known_envs should be the list of annotated environment name objects
        # as returned by app_utils.get_unique_annotated_environment_names, where each
        # object contains these fields: name, short_name, full_name, public_name, foursight_name
        self._known_envs = known_envs
        # Set any green/blue production/staging info.
        for known_env in self._known_envs:
            if self._env_contains(known_env, "blue"):
                known_env["color"] = "blue"
            elif self._env_contains(known_env, "green"):
                known_env["color"] = "green"
            if known_env.get("color"):
                if self._env_contains(known_env, "stage") or self._env_contains(known_env, "staging"):
                    known_env["is_staging"] = True
                else:
                    known_env["is_production"] = True

    def get_known_envs(self) -> list:
        return self._known_envs

    def get_known_envs_count(self) -> int:
        return len(self._known_envs)

    @function_cache
    def get_known_envs_with_gac_names(self) -> list:
        known_envs = copy.deepcopy(self._known_envs)
        for known_env in known_envs:
            gac_name = Gac.get_gac_name(known_env["full_name"])
            if gac_name:
                known_env["gac_name"] = gac_name
        return known_envs

    @staticmethod
    def get_default_env() -> str:
        return os.environ.get("ENV_NAME", Envs._DEFAULT_ENV_PLACHOLDER)

    @function_cache
    def is_known_env(self, env: str) -> bool:
        return self.find_known_env(env) is not None

    @function_cache
    def find_known_env(self, env: str) -> Optional[dict]:
        known_envs = self.get_known_envs_with_gac_names()
        return find_association(known_envs, foursight_name=foursight_env_name(env))

    def is_allowed_env(self, env: str, allowed_envs: list) -> bool:
        if not env or not allowed_envs:
            return False
        for allowed_env in allowed_envs:
            if self.is_same_env(env, allowed_env):
                return True
        return False

    @function_cache
    def is_same_env(self, env_a: str, env_b: str) -> bool:
        return foursight_env_name(env_a) == foursight_env_name(env_b)

    def get_user_auth_info(self, email: str, raise_exception: bool = False) -> Tuple[list, str, str]:
        """
        Returns a tuple containing (in left-right order): the list of known environments,
        i.e the list of annotated environment name objects; the list of allowed environment
        names for the given user/email, via the users store in ElasticSearch; and since we're
        getting the user record anyways, the first/last name of the user, for display only.
        """
        allowed_envs = []
        first_name = None
        last_name = None
        for known_env in self._known_envs:
            try:
                # Note we must lower case the email to find the user. This is because all emails
                # in the database are lowercased; it causes issues with OAuth if we don't do this.
                known_env_name = known_env["full_name"]
                envs = app.core.init_environments(known_env_name)
                connection = app.core.init_connection(known_env_name, envs)
                user = ff_utils.get_metadata('users/' + email.lower(),
                                             key=connection.ff_keys,
                                             add_on="frame=object&datastore=database")
                if self._is_user_allowed_access(user):
                    # Since this is in a loop, for each env, this setup here will end up getting first/last name
                    # from the last env in the loop; doesn't really matter, just pick one set; this is just for
                    # informational/display purposes in the UI.
                    first_name = user.get("first_name")
                    last_name = user.get("last_name")
                    allowed_envs.append(known_env_name)
            except Exception as e:
                if raise_exception:
                    raise
                logger.warning(f"Exception getting allowed envs for {email}: {e}")
        return allowed_envs, first_name, last_name

    @staticmethod
    def _is_user_allowed_access(user: Optional[dict]) -> bool:
        return user and Envs._is_user_in_one_or_more_groups(user, ["admin", "foursight"])

    @staticmethod
    def _is_user_in_one_or_more_groups(user: Optional[dict], allowed_groups: list) -> bool:
        user_groups = user.get("groups") if user else None
        return user_groups and any(allowed_group in user_groups for allowed_group in allowed_groups or [])

    @staticmethod
    def _env_contains(env: dict, value: str, ignore_case: bool = True) -> bool:
        if ignore_case:
            value = value.lower()
            return (value in env["full_name"].lower() or
                    value in env["short_name"].lower() or
                    value in env["public_name"].lower() or
                    value in env["foursight_name"].lower())
        else:
            return (value in env["full_name"] or
                    value in env["short_name"] or
                    value in env["public_name"] or
                    value in env["foursight_name"])

    @staticmethod
    def _env_contained_within(env: dict, value: str) -> bool:
        """
        Returns True iff the given environment (dictionary) is contained or somehow represented
        within the given string value. Originally created for determining (sort of heuristically)
        the environment to which an AWS task definition name should be associated. For example,
        the name "c4-ecs-fourfront-hotseat-stack-FourfrontDeployment-xTDwbIYxIZh7" would belong
        to the "hotseat" environment.
        """
        value = value.lower()
        if "color" in env and env["color"] in ["blue", "green"] and env["color"] in value:
            # Handle situations like this where both blue and green appear, but green appears twice:
            # c4-ecs-blue-green-smaht-production-stack-SmahtgreenDeployment-mIHBLXIQ1pok
            blue_count = value.count("blue")
            green_count = value.count("green")
            if env["color"] == "blue":
                if blue_count > green_count:
                    return True
            elif green_count > blue_count:
                return True
        result = (env["full_name"].lower() in value or
                  env["short_name"].lower() in value or
                  env["public_name"].lower() in value or
                  env["foursight_name"].lower() in value)
        return result

    def get_production_color(self) -> Tuple[Optional[str], Optional[str]]:
        for known_env in self._known_envs:
            if known_env.get("is_production"):
                return (known_env["color"], known_env)
        return (None, None)

    def get_staging_color(self) -> Tuple[Optional[str], Optional[str]]:
        for known_env in self._known_envs:
            if known_env.get("is_staging"):
                return (known_env["color"], known_env)
        return (None, None)

    def get_associated_env(self, name: str) -> Optional[dict]:
        known_envs_with_colors = [env for env in self._known_envs if env.get("color")]
        known_envs_sans_colors = [env for env in self._known_envs if not env.get("color")]
        for known_env in known_envs_with_colors:
            if self._env_contained_within(known_env, name):
                return known_env
        for known_env in known_envs_sans_colors:
            if self._env_contained_within(known_env, name):
                return known_env
        return None



import boto3
import json
from typing import Dict, List, Optional
from dcicutils.ecs_utils import ECSUtils
from dcicutils.misc_utils import get_error_message
from .aws_network import aws_get_security_groups, aws_get_subnets, aws_get_vpcs
from .datetime_utils import convert_datetime_to_utc_datetime_string as datetime_string
from .envs import Envs
from .misc_utils import find_common_prefix
from .portal_access_key_utils import get_portal_access_key_info

# Functions to get AWS cluster and task info with the
# original end purpose of supporting reindexing via Foursight.


def get_aws_ecs_tasks_for_running(envs: Envs, task_definition_type: Optional[str] = None) -> List:

    given_task_definition_type = task_definition_type.lower() if task_definition_type else None
    task_definition_arns = _get_task_definition_arns()
    tasks_for_running = []

    def get_cluster_for_env(clusters: List[Dict], env: Optional[Dict]) -> Optional[str]:
        if not env:
            return None
        for cluster in clusters:
            if envs._env_contained_within(env, cluster):
                return cluster

    def get_vpc() -> Optional[Dict]:
        vpcs = aws_get_vpcs()
        if len(vpcs) == 1:
            vpc = vpcs[0]
        else:
            vpcs = [item for item in vpcs if "main" in (item.get("name") or "").lower()]
            vpc = vpcs[0] if len(vpcs) == 1 else None
        if vpc:
            vpc = {"id": vpc["id"], "name": vpc["name"]}
        return vpc

    def get_security_groups(vpc: Optional[Dict]) -> List[Dict]:
        if not vpc:
            return []
        security_groups = aws_get_security_groups(vpc_id=vpc["id"]) if vpc else []
        security_groups = [item for item in security_groups if "container" in (item.get("name") or "").lower()]
        security_groups = [
            {
                "id": item["id"],
                "name": item["name"],
                "stack": item["stack"]
            }
            for item in security_groups
        ]
        return security_groups

    def get_security_group_for_env(security_groups: List[Dict], env: Optional[Dict]) -> Optional[Dict]:
        if not env:
            return None
        env_specific_security_group = None
        for security_group in security_groups:
            prefix = find_common_prefix([task_definition_arn, security_group["name"], security_group["stack"]])
            if prefix == security_group["stack"]:
                env_specific_security_group = security_group
                break
            elif envs._env_contained_within(env, security_group["name"]):
                env_specific_security_group = security_group
                break
        return env_specific_security_group

    def get_subnets() -> List[Dict]:
        subnets = aws_get_subnets()
        subnets = [item for item in subnets if item.get("type") == "private"]
        return [{"id": subnet["id"], "name": subnet["name"]} for subnet in subnets]

    def get_subnets_for_env(subnets: List[Dict], env: Optional[Dict]) -> List[Dict]:
        subnets_for_env = [item for item in subnets if "main" in (item.get("name") or "").lower()]
        if not subnets_for_env:
            for subnet in subnets:
                if envs._env_contained_within(env, subnet["name"]):
                    subnets_for_env.append(subnet)
        if not subnets_for_env:
            # If none just take all of the (private) subnets.
            subnets_for_env = subnets
        return subnets_for_env

    def add_task_for_running(task: Dict) -> None:
        # Add the given task to the given task list, but make sure we don't have a duplicate,
        # meaning more than one task with the same "name" (i.e. e.g. "deploy") as determined
        # by the _get_task_definition_type function definition, above); and for the same
        # environment, as determined by the get_assocated_env function call, below.
        duplicate_tasks = [existing_task for existing_task in tasks_for_running
                           if existing_task["type"] == task["type"]
                           and existing_task["env"] == task["env"]]
        if duplicate_tasks:
            # Since we are doing this as we go we are
            # guaranteed to get at most only one of these.
            existing_task = duplicate_tasks[0]
            ecs = boto3.client("ecs")
            existing_task_definition = (
                ecs.describe_task_definition(taskDefinition=existing_task["task_definition_arn"])["taskDefinition"])
            this_task_definition = (
                ecs.describe_task_definition(taskDefinition=task["task_definition_arn"])["taskDefinition"])
            existing_task_registered_at = existing_task_definition.get("registeredAt")
            this_task_registered_at = this_task_definition.get("registeredAt")
            if existing_task_registered_at and this_task_registered_at:
                if existing_task_registered_at > this_task_registered_at:
                    # The existing task is newer than this task; skip this one.
                    existing_task["registered_at"] = datetime_string(existing_task_registered_at)
                    if not existing_task.get("duplicates"):
                        existing_task["duplicates"] = []
                    existing_task["duplicates"].append({
                        "task_definition_arn": task["task_definition_arn"],
                        "registered_at": datetime_string(this_task_registered_at)
                     })
                    return
                else:
                    # This task is newer than the existing task; remove the existing one.
                    task["registered_at"] = datetime_string(this_task_registered_at)
                    task["duplicates"] = existing_task.get("duplicates") or []
                    task["duplicates"].append({
                        "task_definition_arn": existing_task["task_definition_arn"],
                        "registered_at": datetime_string(existing_task_registered_at)
                     })
                    tasks_for_running.remove(existing_task)
        tasks_for_running.append(task)

    clusters = _get_cluster_arns()
    vpc = get_vpc()
    security_groups = get_security_groups(vpc)
    subnets = get_subnets()

    for task_definition_arn in task_definition_arns:
        task_definition_type = _get_task_definition_type(task_definition_arn)
        if given_task_definition_type and given_task_definition_type != task_definition_type:
            continue
        task_env = envs.get_associated_env(task_definition_arn)
        task_for_running = {
            "task_definition_arn": task_definition_arn,
            "type": task_definition_type,
            "env": task_env
        }
        if vpc:
            task_for_running["vpc"] = vpc
        # Get the AWS cluster to use for any task run for this particular environment.
        cluster_for_env = get_cluster_for_env(clusters, task_env)
        if cluster_for_env:
            task_for_running["cluster_arn"] = cluster_for_env
        # Get the AWS security groups to use for any task run for this particular environment.
        security_group_for_env = get_security_group_for_env(security_groups, task_env)
        if security_group_for_env:
            task_for_running["security_group"] = security_group_for_env
        # Get the AWS subnets to use for any task run for this particular environment.
        subnets_for_env = get_subnets_for_env(subnets, task_env)
        if subnets_for_env:
            task_for_running["subnets"] = subnets_for_env
        # Add this task to the results (handles "duplicates").
        add_task_for_running(task_for_running)
    return tasks_for_running


def get_aws_ecs_task_running(envs: Envs,
                             cluster_arn: str,
                             task_definition_arn: str,
                             check_other_clusters: bool = True) -> Dict:
    """
    Returns an indication of if the given task definition is currently running.
    Note that if the given task definition ARN is not valid or does not exist,
    then NO indication of this is given; it will just say it is not running.
    """
    task_arns = _get_task_arns(cluster_arn, task_definition_arn)
    task_is_running = len(task_arns) > 0
    # TODO: See if the given task is running in any other cluser
    task_definition_type = _get_task_definition_type(task_definition_arn)
    response = {
        "cluster_arn": cluster_arn,
        "task_definition_arn": task_definition_arn,
        "type": task_definition_type,
        "task_running": task_is_running
    }
    if not task_is_running and check_other_clusters:
        for cluster_arn in [item for item in _get_cluster_arns() if item != cluster_arn]:
            sub_response = get_aws_ecs_task_running(envs,
                                                    cluster_arn=cluster_arn,
                                                    task_definition_arn=task_definition_arn,
                                                    check_other_clusters=False)
            if sub_response["task_running"]:
                sub_response["other_cluster"] = True
                return sub_response
    if task_is_running:
        response["task_running_ids"] = [_get_task_running_id(item) for item in task_arns]
    return response


def get_aws_ecs_task_last_run(envs: Envs, cluster_arn: str, task_definition_arn: str) -> Dict:
    # If this is the deploy task the approximate that last time it was run by using the
    # create date of the Portal Access Key as a proxy for when this task last ran since
    # the entrypoint_deployment.bash script creates this as its last step.
    # This is only separate from get_aws_ecs_task_running for performance
    # and responsiveness response within the (React) UI.
    task_definition_type = _get_task_definition_type(task_definition_arn)
    response = {
        "cluster_arn": cluster_arn,
        "task_definition_arn": task_definition_arn,
        "type": task_definition_type,
    }
    if task_definition_type == "deploy":
        task_env = envs.get_associated_env(task_definition_arn)
        if task_env:
            portal_access_key_info = get_portal_access_key_info(task_env["full_name"], logged_in=True)
            if portal_access_key_info:
                response["task_last_ran_at"] = portal_access_key_info["created_at"]
                response["portal_access_key"] = portal_access_key_info["key"]  # Just FYI
    return response


def get_aws_ecs_tasks_running(cluster_arn: Optional[str] = None,
                              task_definition_type: Optional[str] = None,
                              task_definition_arn: Optional[str] = None) -> List[Dict]:
    """
    Returns a list of all tasks running within the given cluster (ARN).
    This list is groups by task definition (ARN), which each containing
    the list of task IDs of associated tasks running.
    """
    if not cluster_arn:
        return get_aws_ecs_tasks_running_across_clusters(task_definition_type=task_definition_type,
                                                         task_definition_arn=task_definition_arn)
    response = []
    ecs = boto3.client("ecs")
    task_arns = ecs.list_tasks(cluster=cluster_arn).get("taskArns")
    if task_arns:
        given_task_definition_type = task_definition_type
        given_task_definition_arn = task_definition_arn
        tasks = ecs.describe_tasks(cluster=cluster_arn, tasks=task_arns).get("tasks")
        for task in tasks:
            task_definition_arn = _shortened_task_definition_arn(task.get("taskDefinitionArn"))
            if given_task_definition_arn and given_task_definition_arn != task_definition_arn:
                continue
            task_definition_type = _get_task_definition_type(task_definition_arn)
            if given_task_definition_type and given_task_definition_type != task_definition_type:
                continue
            task = {
                "id": _get_task_running_id(task.get("taskArn")),
                "started_at": datetime_string(task.get("startedAt") or task.get("createdAt"))
            }
            if not given_task_definition_arn:
                task["cluster_arn"] = cluster_arn
                existing_task = [item for item in response if item["task_definition_arn"] == task_definition_arn]
            else:
                existing_task = [item for item in response if item["cluster_arn"] == cluster_arn]
            if existing_task:
                existing_task[0]["tasks"].append(task)
            elif not given_task_definition_arn:
                response.append({"task_definition_arn": task_definition_arn,
                                 "type": _get_task_definition_type(task_definition_arn),
                                 "tasks": [task]})
            else:
                response.append({"cluster_arn": cluster_arn,
                                 "task_definition_arn": given_task_definition_arn,
                                 "type": _get_task_definition_type(task_definition_arn),
                                 "tasks": [task]})
    return response


def get_aws_ecs_tasks_running_across_clusters(task_definition_type: Optional[str] = None,
                                              task_definition_arn: Optional[str] = None) -> List[Dict]:
    """
    Returns tasks running across any/all clusters.
    If a task_definition_arn is given the the response is will contain only tasks which are
    instances of that task_definition_arn grouped, and the response is grouped by by cluster_arn.
    If a task_definition_arn is not given the the response is grouped by task_definition_arn.
    """
    response = []
    cluster_arns = [item for item in _get_cluster_arns()]
    for cluster_arn in cluster_arns:
        tasks = get_aws_ecs_tasks_running(cluster_arn, task_definition_type, task_definition_arn)
        for task in tasks:
            if not task_definition_arn:
                existing_task = (
                    [item for item in response if item["task_definition_arn"] == task["task_definition_arn"]])
            else:
                existing_task = [item for item in response if item["cluster_arn"] == task["cluster_arn"]]
            if existing_task:
                existing_task[0]["tasks"].extend(task["tasks"])
            else:
                response.append(task)
    return response


def aws_ecs_run_task(cluster_arn: str, task_definition_arn: str, args: Dict) -> Dict:
    subnets = args.get("subnets")
    security_group = args.get("security_group")
    subnets = [item["id"] for item in subnets]
    security_group = security_group["id"]
    # TODO: The dcicutils.ecs_utils.run_ecs_task function does not specify specify
    # container launchType (FARGATE), for some reason, one of which is required.
    # ecs = ECSUtils()
    # response = ecs.run_ecs_task(
    #     cluster_name=cluster_arn,
    #     task_definition_type=task_definition_arn,
    #     subnet=subnets,  # TODO
    #     security_group=security_group
    # )
    try:
        response = {
            "cluster_arn": cluster_arn,
            "task_definition_arn": task_definition_arn,
            "security_group": security_group,
            "subnets": subnets
        }
        ecs = boto3.client("ecs")
        run_task_response = ecs.run_task(
            launchType="FARGATE",
            count=1,
            cluster=cluster_arn,
            taskDefinition=task_definition_arn,
            networkConfiguration={"awsvpcConfiguration": {"subnets": subnets, "securityGroups": [security_group]}}
        )
        response["task_running_id"] = _get_task_running_id(run_task_response.get("tasks", [{}])[0].get("taskArn"))
        response["response"] = json.loads(json.dumps(run_task_response, default=str))
    except Exception as e:
        response["error"] = get_error_message(e)
    return response


def _get_cluster_arns() -> List[str]:
    ecs = ECSUtils()
    return sorted([_shortened_arn(item) for item in ecs.list_ecs_clusters()])


def _get_task_definition_arns() -> List[str]:
    ecs = ECSUtils()
    return sorted(list(set([_shortened_task_definition_arn(item) for item in ecs.list_ecs_tasks()])))


def _get_task_arns(cluster_arn: str, task_definition_arn: str) -> List[str]:
    ecs = boto3.client("ecs")
    return ecs.list_tasks(cluster=cluster_arn, family=task_definition_arn).get("taskArns")


def _shortened_arn(arn: str) -> str:
    arn_parts = arn.split("/", 1) if arn else []
    return arn_parts[1] if len(arn_parts) > 1 else arn


def _shortened_task_definition_arn(task_definition_arn: str) -> str:
    """
    Given something like this:
    - arn:aws:ecs:us-east-1:466564410312:task-definition/c4-ecs-cgap-supertest-stack-CGAPDeployment-of2dr96JX1ds:1
    this function would return this:
    - c4-ecs-cgap-supertest-stack-CGAPDeployment-of2dr96JX1ds
    """
    task_definition_arn = _shortened_arn(task_definition_arn)
    arn_parts = task_definition_arn.rsplit(":", 1) if task_definition_arn else []
    return arn_parts[0] if len(arn_parts) > 1 else task_definition_arn


def _get_task_definition_type(task_definition_arn: str) -> str:
    if "deploy" in task_definition_arn.lower():
        if "initial" in task_definition_arn.lower():
            return "deploy_initial"
        else:
            return "deploy"
    elif "index" in task_definition_arn.lower():
        return "indexer"
    elif "ingest" in task_definition_arn.lower():
        return "ingester"
    elif "portal" in task_definition_arn.lower():
        return "portal"
    else:
        return task_definition_arn


def _get_task_running_id(task_arn: str) -> str:
    return task_arn.split("/")[-1] if "/" in task_arn else task_arn



from chalice import Chalice
from json import loads as load_json
from typing import Optional


app = Chalice(app_name='foursight-core')
app.request = lambda: app.current_request.to_dict()
app.request_args = lambda: app.request().get("query_params", {}) or {}
app.request_arg = lambda name, fallback = None: app.request_args().get(name, fallback)
app.request_arg_int = lambda name, fallback = 0: parse_int(app.request_arg(name), fallback)
app.request_arg_bool = lambda name, fallback = False: parse_bool(app.request_arg(name), fallback)
app.request_body = lambda: load_json(app.current_request.raw_body.decode())
app.request_method = lambda: app.current_request.method.upper()


def parse_int(value: Optional[str], fallback: int = 0) -> int:
    try:
        return int(value) if value else fallback
    except ValueError:
        return fallback


def parse_bool(value: Optional[str], fallback: int = 0) -> int:
    try:
        return value.lower() == "true" if value else fallback
    except ValueError:
        return fallback



import datetime
from typing import Any, Optional
from http.cookies import SimpleCookie
from dcicutils.misc_utils import str_to_bool
from .datetime_utils import (
    convert_time_t_to_datetime,
    convert_utc_datetime_to_cookie_expires_format,
    EPOCH
)
from .misc_utils import is_running_locally


def read_cookie(request: dict, cookie_name: str) -> Optional[str]:
    """
    Returns the value of the cookie of the given name from within the given request,
    or None if it does not exist.
    """
    if not cookie_name or not request:
        return ""
    simple_cookies = _read_cookies(request)
    cookie_morsel = simple_cookies.get(cookie_name) if simple_cookies else None
    return cookie_morsel.value if cookie_morsel else None


def read_cookie_bool(request: dict, cookie_name: str) -> bool:
    """
    Returns the value of the cookie of the given name from within the given request,
    as a bool (per dcicutils.str_to_bool), or False if the cookie does not exist
    or is not a value True specifier.
    """
    result = str_to_bool(read_cookie(request, cookie_name))
    return True if result else False


def read_cookie_int(request: dict, cookie_name: str, default: int = 0) -> int:
    """
    Returns the value of the cookie of the given name from within the given request,
    as an integer, or the given default value (which defaults to 0) if the cookie
    does not exist or is not a value a valid integer.
    """
    value = read_cookie(request, cookie_name)
    if not value:
        return default
    if value.startswith("-"):
        value = value[1:]
        multiplier = -1
    else:
        multiplier = 1
    if not value.isdigit():
        return default
    return int(value) * multiplier


def _read_cookies(request: dict) -> Optional[SimpleCookie]:
    """
    Returns the cookies from the given request as a SimpleCookie object.
    """
    if not request:
        return None
    cookies = request.get("headers", {}).get("cookie")
    if not cookies:
        return None
    simple_cookies = SimpleCookie()
    simple_cookies.load(cookies)
    return simple_cookies


def create_set_cookie_string(request: dict, name: str, value: Optional[str],
                             domain: str, path: str = "/", expires: Optional[Any] = None,
                             http_only: bool = False) -> str:
    """
    Returns a string suitable for an HTTP response to set a cookie for this given cookie info.
    If the given expires arg is "now" then the expiration time for the cookie will be set
    to the epoch (i.e. 1970-01-01 00:00:00) indicating it has expired; used to delete cookie.
    """
    if not name or not request:
        return ""
    cookie = name + "=" + (value if value else "") + ";"
    if domain and not is_running_locally(request):
        # N.B. When running on localhost cookies cannot be set unless we leave off the domain entirely.
        # https://stackoverflow.com/questions/1134290/cookies-on-localhost-with-explicit-domain
        cookie += f" Domain={domain};"
    if not path:
        path = "/"
    cookie += f" Path={path};"
    if expires:
        if isinstance(expires, datetime.datetime):
            expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        elif isinstance(expires, int):
            expires = convert_time_t_to_datetime(expires)
            expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        elif isinstance(expires, str):
            if expires.lower() == "now":
                expires = f"Expires={convert_utc_datetime_to_cookie_expires_format(EPOCH)}"
        else:
            expires = None
        if expires:
            cookie += f" Expires={expires};"
    if http_only:
        cookie += " HttpOnly;"
    return cookie


def create_delete_cookie_string(request: dict, name: str, domain: str, path: str = "/") -> str:
    return create_set_cookie_string(request, name=name, value=None, domain=domain, path=path, expires="now")
